import logging
import os
from typing import Any, Callable, Dict, List, Optional, Type, Union

import dask.dataframe as dd
import pandas as pd
from distributed import Client
from triad.collections import Schema
from triad.collections.dict import IndexedOrderedDict, ParamDict
from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid
from triad.utils.pandas_like import PandasUtils
from triad.utils.threading import RunOnce
from triad.utils.io import makedirs
from fugue import StructuredRawSQL
from fugue.collections.partition import (
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from fugue.dataframe import (
    AnyDataFrame,
    DataFrame,
    DataFrames,
    LocalDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.utils import get_join_schemas
from fugue.exceptions import FugueBug
from fugue.execution.execution_engine import ExecutionEngine, MapEngine, SQLEngine
from fugue.execution.native_execution_engine import NativeExecutionEngine
from fugue_dask._constants import FUGUE_DASK_DEFAULT_CONF
from fugue_dask._io import load_df, save_df
from fugue_dask._utils import (
    DASK_UTILS,
    DaskUtils,
    even_repartition,
    hash_repartition,
    rand_repartition,
)
from fugue_dask.dataframe import DaskDataFrame

from ._constants import FUGUE_DASK_USE_ARROW

_DASK_PARTITION_KEY = "__dask_partition_key__"


class DaskSQLEngine(SQLEngine):
    """Dask-sql implementation."""

    @property
    def dialect(self) -> Optional[str]:
        return "trino"

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return to_dask_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return True

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        try:
            from dask_sql import Context
        except ImportError:  # pragma: no cover
            raise ImportError(
                "dask-sql is not installed. "
                "Please install it with `pip install dask-sql`"
            )
        ctx = Context()
        _dfs: Dict[str, dd.DataFrame] = {k: self._to_safe_df(v) for k, v in dfs.items()}
        sql = statement.construct(dialect=self.dialect, log=self.log)
        res = ctx.sql(
            sql,
            dataframes=_dfs,
            config_options={"sql.identifier.case_sensitive": True},
        )
        return DaskDataFrame(res)

    def _to_safe_df(self, df: DataFrame) -> dd.DataFrame:
        df = self.to_df(df)
        return df.native.astype(
            df.schema.to_pandas_dtype(use_extension_types=True, use_arrow_dtype=False)
        )


class DaskMapEngine(MapEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return DaskExecutionEngine

    @property
    def is_distributed(self) -> bool:
        return True

    def map_dataframe(  # noqa: C901
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        presort = partition_spec.get_sorts(
            df.schema, with_partition_keys=partition_spec.algo == "coarse"
        )
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        output_schema = Schema(output_schema)
        output_dtypes = output_schema.to_pandas_dtype(
            use_extension_types=True, use_arrow_dtype=FUGUE_DASK_USE_ARROW
        )
        input_schema = df.schema
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _fix_dask_bug(pdf: pd.DataFrame) -> pd.DataFrame:
            assert_or_throw(
                pdf.shape[1] == len(input_schema),
                FugueBug(
                    "partitioned dataframe has different number of columns: "
                    f"{pdf.columns} vs {input_schema}"
                ),
            )
            return pdf

        def _core_map(pdf: pd.DataFrame) -> pd.DataFrame:
            if len(partition_spec.presort) > 0:
                pdf = pdf.sort_values(presort_keys, ascending=presort_asc)
            input_df = PandasDataFrame(
                pdf.reset_index(drop=True), input_schema, pandas_df_wrapper=True
            )
            if on_init_once is not None:
                on_init_once(0, input_df)
            cursor.set(lambda: input_df.peek_array(), 0, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_pandas()[output_schema.names]

        def _map(pdf: pd.DataFrame) -> pd.DataFrame:
            if pdf.shape[0] == 0:
                return PandasDataFrame([], output_schema).as_pandas()
            pdf = pdf.reset_index(drop=True)
            pdf = _fix_dask_bug(pdf)
            res = _core_map(pdf)
            return res.astype(output_dtypes)

        def _gp_map(pdf: pd.DataFrame) -> pd.DataFrame:
            if pdf.shape[0] == 0:  # pragma: no cover
                return PandasDataFrame([], output_schema).as_pandas()
            pdf = pdf.reset_index(drop=True)
            pdf = _fix_dask_bug(pdf)
            pu = PandasUtils()
            res = pu.safe_groupby_apply(pdf, partition_spec.partition_by, _core_map)
            return res.astype(output_dtypes)

        df = self.to_df(df)
        pdf = self.execution_engine.repartition(df, partition_spec)
        if len(partition_spec.partition_by) == 0:
            result = pdf.native.map_partitions(_map, meta=output_dtypes)  # type: ignore
        else:
            if partition_spec.algo == "default":
                result = df.native.groupby(
                    partition_spec.partition_by,
                    sort=False,
                    group_keys=False,
                    dropna=False,
                ).apply(_map, meta=output_dtypes)
            elif partition_spec.algo == "coarse":
                result = pdf.native.map_partitions(  # type: ignore
                    _map, meta=output_dtypes
                )
            else:
                result = pdf.native.map_partitions(  # type: ignore
                    _gp_map, meta=output_dtypes
                )
        return DaskDataFrame(result, output_schema)


class DaskExecutionEngine(ExecutionEngine):
    """The execution engine based on `Dask <https://docs.dask.org/>`_.

    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param dask_client: Dask distributed client, defaults to None. If None, then it
      will try to get the current active global client. If there is no active client,
      it will create and use a global `Client(processes=True)`
    :param conf: |ParamsLikeObject| defaults to None, read |FugueConfig| to
      learn Fugue specific options

    .. note::

        You should setup Dask single machine or distributed environment in the
        :doc:`common <dask:setup>` way.
        Before initializing :class:`~.DaskExecutionEngine`
    """

    def __init__(self, dask_client: Optional[Client] = None, conf: Any = None):
        p = ParamDict(FUGUE_DASK_DEFAULT_CONF)
        p.update(ParamDict(conf))
        super().__init__(p)
        self._log = logging.getLogger()
        self._client = DASK_UTILS.get_or_create_client(dask_client)
        self._native = NativeExecutionEngine(conf=conf)

    def __repr__(self) -> str:
        return "DaskExecutionEngine"

    @property
    def is_distributed(self) -> bool:
        return True

    @property
    def dask_client(self) -> Client:
        """The Dask Client associated with this engine"""
        return self._client

    @property
    def log(self) -> logging.Logger:
        return self._log

    def create_default_sql_engine(self) -> SQLEngine:
        return DaskSQLEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return DaskMapEngine(self)

    def get_current_parallelism(self) -> int:
        res = dict(self.dask_client.nthreads())
        return sum(res.values())

    @property
    def pl_utils(self) -> DaskUtils:
        """Pandas-like dataframe utils"""
        return DaskUtils()

    def to_df(self, df: Any, schema: Any = None) -> DaskDataFrame:
        """Convert a data structure to :class:`~.fugue_dask.dataframe.DaskDataFrame`

        :param data: :class:`~.fugue.dataframe.dataframe.DataFrame`,
          :class:`dask:dask.dataframe.DataFrame`,
          pandas DataFrame or list or iterable of arrays
        :param schema: |SchemaLikeObject|, defaults to None.
        :return: engine compatible dataframe

        .. note::

            * if the input is already :class:`~.fugue_dask.dataframe.DaskDataFrame`,
              it should return itself
            * For list or iterable of arrays, ``schema`` must be specified
            * When ``schema`` is not None, a potential type cast may happen to ensure
              the dataframe's schema.
            * all other methods in the engine can take arbitrary dataframes and
              call this method to convert before doing anything
        """

        return to_dask_engine_df(df, schema)

    def repartition(
        self, df: DataFrame, partition_spec: PartitionSpec
    ) -> DaskDataFrame:
        df = self.to_df(df)
        if partition_spec.empty:
            return df
        if len(partition_spec.partition_by) > 0 and partition_spec.algo == "default":
            return df

        p = partition_spec.get_num_partitions(
            **{
                KEYWORD_ROWCOUNT: lambda: df.persist().count(),  # type: ignore
                KEYWORD_PARALLELISM: lambda: self.get_current_parallelism(),
            }
        )
        if partition_spec.algo == "default":
            ddf: dd.DataFrame = (
                df.native.repartition(npartitions=p) if p > 0 else df.native
            )
        elif partition_spec.algo in ["hash", "coarse"]:
            ddf = hash_repartition(
                df.native,
                num=p if p > 0 else self.get_current_parallelism() * 2,
                cols=partition_spec.partition_by,
            )
        elif partition_spec.algo == "even":
            ddf = even_repartition(df.native, num=p, cols=partition_spec.partition_by)
        elif partition_spec.algo == "rand":
            ddf = rand_repartition(
                df.native,
                num=p if p > 0 else self.get_current_parallelism() * 2,
                cols=partition_spec.partition_by,
                seed=0,
            )
        else:  # pragma: no cover
            raise NotImplementedError(
                partition_spec.algo + " partitioning is not supported"
            )
        if ddf is None or df.native is ddf:
            return df
        return DaskDataFrame(ddf, schema=df.schema, type_safe=False)

    def broadcast(self, df: DataFrame) -> DataFrame:
        return self.to_df(df)

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        res = self.to_df(df)
        res.reset_metadata(df.metadata)
        return res.persist()

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        # Dask joins on different types such as int64 vs Int64 can occasionally fail
        # so we need to cast to the same type
        ndf1 = self.to_df(df1).native
        ntp1 = ndf1.dtypes[key_schema.names].to_dict()
        ndf2 = self.to_df(df2).native
        ntp2 = ndf2.dtypes[key_schema.names].to_dict()
        if ntp1 != ntp2:
            ntp = key_schema.to_pandas_dtype(
                use_extension_types=True, use_arrow_dtype=FUGUE_DASK_USE_ARROW
            )
            if ntp1 != ntp:
                ndf1 = ndf1.astype(ntp)
            if ntp2 != ntp:
                ndf2 = ndf2.astype(ntp)

        d = self.pl_utils.join(ndf1, ndf2, join_type=how, on=key_schema.names)
        return DaskDataFrame(d, output_schema, type_safe=False)

    def union(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.union(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema, type_safe=False)

    def subtract(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("EXCEPT ALL for DaskExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.except_df(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema, type_safe=False)

    def intersect(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("INTERSECT ALL for DaskExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.intersect(
            self.to_df(df1).native, self.to_df(df2).native, unique=distinct
        )
        return DaskDataFrame(d, df1.schema, type_safe=False)

    def distinct(self, df: DataFrame) -> DataFrame:
        d = self.pl_utils.drop_duplicates(self.to_df(df).native)
        return DaskDataFrame(d, df.schema, type_safe=False)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: List[str] = None,
    ) -> DataFrame:
        kw: Dict[str, Any] = dict(how=how)
        if thresh is not None:
            kw["thresh"] = thresh
        if subset is not None:
            kw["subset"] = subset
        if how == "any" and thresh is not None:
            del kw["how"]  # to deal with a dask logic flaw
        d = self.to_df(df).native.dropna(**kw)
        return DaskDataFrame(d, df.schema, type_safe=False)

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            ValueError("fillna value can not be a list or None"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                ValueError(
                    "fillna dict can not contain None and needs at least one value"
                ),
            )
            mapping = value
        else:
            # If subset is none, apply to all columns
            subset = subset or df.columns
            mapping = {col: value for col in subset}
        d = self.to_df(df).native.fillna(mapping)
        return DaskDataFrame(d, df.schema, type_safe=False)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )
        # TODO: dask does not support sample by number of rows
        d = self.to_df(df).native.sample(
            n=n, frac=frac, replace=replace, random_state=seed
        )
        return DaskDataFrame(d, df.schema, type_safe=False)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        d = self.to_df(df).native
        meta = [(d[x].name, d[x].dtype) for x in d.columns]

        if presort:
            presort = parse_presort_exp(presort)
        # Use presort over partition_spec.presort if possible
        _presort: IndexedOrderedDict = presort or partition_spec.presort

        def _partition_take(partition, n, presort):
            assert_or_throw(
                partition.shape[1] == len(meta),
                FugueBug("hitting the dask bug where partition keys are lost"),
            )
            if len(presort.keys()) > 0 and len(partition) > 1:
                partition = partition.sort_values(
                    list(presort.keys()),
                    ascending=list(presort.values()),
                    na_position=na_position,
                )
            return partition.head(n)

        if len(partition_spec.partition_by) == 0:
            if len(_presort.keys()) == 0:
                d = d.head(n)
            else:
                # Use the default partition
                d = (
                    d.map_partitions(_partition_take, n, _presort, meta=meta)
                    .reset_index(drop=True)
                    .compute()
                )
                # compute() brings this to Pandas so we can use pandas
                d = d.sort_values(
                    list(_presort.keys()),
                    ascending=list(_presort.values()),
                    na_position=na_position,
                ).head(n)

        else:
            d = (
                d.groupby(partition_spec.partition_by, dropna=False)
                .apply(_partition_take, n=n, presort=_presort, meta=meta)
                .reset_index(drop=True)
            )

        return DaskDataFrame(d, df.schema, type_safe=False)

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DaskDataFrame:
        return self.to_df(
            load_df(path, format_hint=format_hint, columns=columns, **kwargs)
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if force_single:
            self._native.save_df(
                df,
                path=path,
                format_hint=format_hint,
                mode=mode,
                partition_spec=partition_spec,
                force_single=force_single,
                **kwargs,
            )
        else:
            if not partition_spec.empty:
                kwargs["partition_on"] = partition_spec.partition_by
            makedirs(os.path.dirname(path), exist_ok=True)
            df = self.to_df(df)
            save_df(df, path, format_hint=format_hint, mode=mode, **kwargs)


def to_dask_engine_df(df: Any, schema: Any = None) -> DaskDataFrame:
    """Convert a data structure to :class:`~.fugue_dask.dataframe.DaskDataFrame`

    :param data: :class:`~.fugue.dataframe.dataframe.DataFrame`,
      :class:`dask:dask.dataframe.DataFrame`,
      pandas DataFrame or list or iterable of arrays
    :param schema: |SchemaLikeObject|, defaults to None.
    :return: engine compatible dataframe

    .. note::

        * if the input is already :class:`~fugue_dask.dataframe.DaskDataFrame`,
          it should return itself
        * For list or iterable of arrays, ``schema`` must be specified
        * When ``schema`` is not None, a potential type cast may happen to ensure
          the dataframe's schema.
        * all other methods in the engine can take arbitrary dataframes and
          call this method to convert before doing anything
    """

    if isinstance(df, DataFrame):
        assert_or_throw(
            schema is None,
            ValueError("schema must be None when df is a DataFrame"),
        )
        if isinstance(df, DaskDataFrame):
            return df
        if isinstance(df, PandasDataFrame):
            res = DaskDataFrame(df.native, df.schema)
        else:
            res = DaskDataFrame(df.as_array(type_safe=True), df.schema)
        res.reset_metadata(df.metadata)
        return res
    return DaskDataFrame(df, schema)



from typing import Any

import dask.dataframe as dd
from dask.distributed import Client

from fugue import DataFrame
from fugue.dev import (
    DataFrameParam,
    ExecutionEngineParam,
    fugue_annotated_param,
    is_pandas_or,
)
from fugue.plugins import (
    as_fugue_dataset,
    infer_execution_engine,
    parse_execution_engine,
)
from fugue_dask._utils import DASK_UTILS
from fugue_dask.dataframe import DaskDataFrame
from fugue_dask.execution_engine import DaskExecutionEngine


@infer_execution_engine.candidate(
    lambda objs: is_pandas_or(objs, (dd.DataFrame, DaskDataFrame))
)
def _infer_dask_client(objs: Any) -> Any:
    return DASK_UTILS.get_or_create_client()


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, dd.DataFrame))
def _dask_as_fugue_df(df: dd.DataFrame, **kwargs: Any) -> DaskDataFrame:
    return DaskDataFrame(df, **kwargs)


@parse_execution_engine.candidate(
    lambda engine, conf, **kwargs: isinstance(engine, Client),
    priority=4,  # TODO: this is to overwrite dask-sql fugue integration
)
def _parse_dask_client(engine: Client, conf: Any, **kwargs: Any) -> DaskExecutionEngine:
    return DaskExecutionEngine(dask_client=engine, conf=conf)


@parse_execution_engine.candidate(
    lambda engine, conf, **kwargs: isinstance(engine, str) and engine == "dask",
    priority=4,  # TODO: this is to overwrite dask-sql fugue integration
)
def _parse_dask_str(engine: str, conf: Any, **kwargs: Any) -> DaskExecutionEngine:
    return DaskExecutionEngine(conf=conf)


@fugue_annotated_param(DaskExecutionEngine)
class _DaskExecutionEngineParam(ExecutionEngineParam):
    pass


@fugue_annotated_param(dd.DataFrame)
class _DaskDataFrameParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, DaskExecutionEngine)
        return ctx.to_df(df).native

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, dd.DataFrame)
        assert isinstance(ctx, DaskExecutionEngine)
        return ctx.to_df(output, schema=schema)

    def count(self, df: DataFrame) -> int:  # pragma: no cover
        raise NotImplementedError("not allowed")



from typing import Any, Dict

import dask
import pandas as pd
import pyarrow as pa
from packaging import version

FUGUE_DASK_CONF_DEFAULT_PARTITIONS = "fugue.dask.default.partitions"
FUGUE_DASK_DEFAULT_CONF: Dict[str, Any] = {FUGUE_DASK_CONF_DEFAULT_PARTITIONS: -1}
FUGUE_DASK_USE_ARROW = (
    hasattr(pd, "ArrowDtype")
    and version.parse(dask.__version__) >= version.parse("2023.2")
    and version.parse(pa.__version__) >= version.parse("7")
    and version.parse(pd.__version__) >= version.parse("2")
)



from typing import Any, Callable

import ibis
from pyspark.sql import DataFrame as PySparkDataFrame
from triad.utils.assertion import assert_or_throw

from fugue import DataFrame, DataFrames, ExecutionEngine
from fugue_ibis import IbisTable
from fugue_ibis._utils import to_schema
from fugue_ibis.execution.ibis_engine import IbisEngine, parse_ibis_engine
from fugue_spark.dataframe import SparkDataFrame
from fugue_spark.execution_engine import SparkExecutionEngine


class SparkIbisEngine(IbisEngine):
    def __init__(self, execution_engine: ExecutionEngine) -> None:
        assert_or_throw(
            isinstance(execution_engine, SparkExecutionEngine),
            lambda: ValueError(
                f"SparkIbisEngine must use SparkExecutionEngine ({execution_engine})"
            ),
        )
        super().__init__(execution_engine)

    def select(
        self, dfs: DataFrames, ibis_func: Callable[[ibis.BaseBackend], IbisTable]
    ) -> DataFrame:
        for k, v in dfs.items():
            self.execution_engine.register(v, k)  # type: ignore
        con = ibis.pyspark.connect(self.execution_engine.spark_session)  # type: ignore
        expr = ibis_func(con)
        schema = to_schema(expr.schema())
        result = expr.compile()
        assert_or_throw(
            isinstance(result, PySparkDataFrame),
            lambda: ValueError(f"result must be a PySpark DataFrame ({type(result)})"),
        )
        return SparkDataFrame(result, schema=schema)


@parse_ibis_engine.candidate(
    lambda obj, *args, **kwargs: isinstance(obj, SparkExecutionEngine)
)
def _spark_to_ibis_engine(obj: Any, engine: ExecutionEngine) -> IbisEngine:
    return SparkIbisEngine(engine)



from abc import ABC, abstractmethod

from fugue.dataframe import DataFrame, DataFrames
from fugue.extensions.context import ExtensionContext


class Processor(ExtensionContext, ABC):
    """The interface to process one or multiple incoming dataframes and return one
    DataFrame. For example dropping a column of df should be a type of Processor.
    Processor is task level extension, running on driver, and execution engine aware.

    To implement this class, you should not have ``__init__``, please directly implement
    the interface functions.

    .. note::

      Before implementing this class, do you really need to implement this
      interface? Do you know the interfaceless feature of Fugue? Implementing Processor
      is commonly unnecessary. You can choose the interfaceless approach which may
      decouple your code from Fugue.

    .. seealso::

      Please read
      :doc:`Processor Tutorial <tutorial:tutorials/extensions/processor>`
    """

    @abstractmethod
    def process(self, dfs: DataFrames) -> DataFrame:  # pragma: no cover
        """Process the collection of dataframes on driver side

        .. note::

          * It runs on driver side
          * The dataframes are not necessarily local, for example a SparkDataFrame
          * It is engine aware, you can put platform dependent code in it (for example
            native pyspark code) but by doing so your code may not be portable. If you
            only use the functions of the general ExecutionEngine, it's still portable.

        :param dfs: dataframe collection to process
        :return: the result dataframe
        """
        raise NotImplementedError



import os
import pathlib
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from triad.collections.dict import ParamDict
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.io import join, url_to_fs
from triad.utils.pandas_like import PD_UTILS

from fugue.dataframe import LocalBoundedDataFrame, LocalDataFrame, PandasDataFrame


class FileParser(object):
    def __init__(self, path: str, format_hint: Optional[str] = None):
        self._orig_format_hint = format_hint
        self._has_glob = "*" in path or "?" in path
        self._raw_path = path
        self._fs, self._fs_path = url_to_fs(path)
        if not self.is_local:
            self._path = self._fs.unstrip_protocol(self._fs_path)
        else:
            self._path = os.path.abspath(self._fs._strip_protocol(path))

        if format_hint is None or format_hint == "":
            for k, v in _FORMAT_MAP.items():
                if self.suffix.endswith(k):
                    self._format = v
                    return
            raise NotImplementedError(f"{self.suffix} is not supported")
        else:
            assert_or_throw(
                format_hint in _FORMAT_MAP.values(),
                NotImplementedError(f"{format_hint} is not supported"),
            )
            self._format = format_hint

    def assert_no_glob(self) -> "FileParser":
        assert_or_throw(not self.has_glob, f"{self.raw_path} has glob pattern")
        return self

    @property
    def has_glob(self):
        return self._has_glob

    @property
    def is_local(self):
        return isinstance(self._fs, LocalFileSystem)

    def join(self, path: str, format_hint: Optional[str] = None) -> "FileParser":
        if not self.has_glob:
            _path = join(self.path, path)
        else:
            _path = join(self.parent, path)
        return FileParser(_path, format_hint or self._orig_format_hint)

    @property
    def parent(self) -> str:
        return self._fs.unstrip_protocol(self._fs._parent(self._fs_path))

    @property
    def path(self) -> str:
        return self._path

    @property
    def raw_path(self) -> str:
        return self._raw_path

    @property
    def suffix(self) -> str:
        return "".join(pathlib.Path(self.raw_path.lower()).suffixes)

    @property
    def file_format(self) -> str:
        return self._format

    def make_parent_dirs(self) -> None:
        self._fs.makedirs(self._fs._parent(self._fs_path), exist_ok=True)

    def find_all(self) -> Iterable["FileParser"]:
        if self.has_glob:
            for x in self._fs.glob(self._fs_path):
                yield FileParser(self._fs.unstrip_protocol(x))
        else:
            yield self

    def open(self, *args: Any, **kwargs: Any) -> Any:
        self.assert_no_glob()
        return self._fs.open(self._fs_path, *args, **kwargs)


def load_df(
    uri: Union[str, List[str]],
    format_hint: Optional[str] = None,
    columns: Any = None,
    fs: Optional[AbstractFileSystem] = None,
    **kwargs: Any,
) -> LocalBoundedDataFrame:
    if isinstance(uri, str):
        fp = [FileParser(uri, format_hint)]
    else:
        fp = [FileParser(u, format_hint) for u in uri]
    dfs: List[pd.DataFrame] = []
    schema: Any = None
    for f in _get_single_files(fp, fs):
        df, schema = _FORMAT_LOAD[f.file_format](f.assert_no_glob(), columns, **kwargs)
        dfs.append(df)
    return PandasDataFrame(pd.concat(dfs), schema)


def save_df(
    df: LocalDataFrame,
    uri: str,
    format_hint: Optional[str] = None,
    mode: str = "overwrite",
    fs: Optional[AbstractFileSystem] = None,
    **kwargs: Any,
) -> None:
    assert_or_throw(
        mode in ["overwrite", "error"], NotImplementedError(f"{mode} is not supported")
    )
    p = FileParser(uri, format_hint).assert_no_glob()
    if fs is None:
        fs, _ = url_to_fs(uri)
    if fs.exists(uri):
        assert_or_throw(mode == "overwrite", FileExistsError(uri))
        try:
            fs.rm(uri, recursive=True)
        except Exception:  # pragma: no cover
            pass
    _FORMAT_SAVE[p.file_format](df, p, **kwargs)


def _get_single_files(
    fp: Iterable[FileParser], fs: Optional[AbstractFileSystem]
) -> Iterable[FileParser]:
    for f in fp:
        yield from f.find_all()


def _save_parquet(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    PD_UTILS.to_parquet_friendly(
        df.as_pandas(), partition_cols=kwargs.get("partition_cols", [])
    ).to_parquet(
        p.assert_no_glob().path,
        **{
            "engine": "pyarrow",
            "schema": df.schema.pa_schema,
            **kwargs,
        },
    )


def _load_parquet(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    if columns is None:
        pdf = pd.read_parquet(p.path, **{"engine": "pyarrow", **kwargs})
        return pdf, None
    if isinstance(columns, list):  # column names
        pdf = pd.read_parquet(
            p.path, columns=columns, **{"engine": "pyarrow", **kwargs}
        )
        return pdf, None
    schema = Schema(columns)
    pdf = pd.read_parquet(
        p.path, columns=schema.names, **{"engine": "pyarrow", **kwargs}
    )
    return pdf, schema


def _save_csv(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    with p.open("w") as f:
        df.as_pandas().to_csv(f, **{"index": False, "header": False, **kwargs})


def _safe_load_csv(p: FileParser, **kwargs: Any) -> pd.DataFrame:
    def load_dir() -> pd.DataFrame:
        dfs: List[pd.DataFrame] = []
        for _p in p.join("*.csv").find_all():  # type: ignore
            with _p.open("r") as f:
                dfs.append(pd.read_csv(f, **kwargs))
        return pd.concat(dfs)

    try:
        with p.open("r") as f:
            return pd.read_csv(f, **kwargs)
    except IsADirectoryError:
        return load_dir()
    except pd.errors.ParserError:  # pragma: no cover
        return load_dir()
    except PermissionError:  # pragma: no cover
        # for windows
        return load_dir()


def _load_csv(  # noqa: C901
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    kw = ParamDict(kwargs)
    infer_schema = kw.get("infer_schema", False)
    if infer_schema and columns is not None and not isinstance(columns, list):
        raise ValueError("can't set columns as a schema when infer schema is true")
    if not infer_schema:
        kw["dtype"] = object
    if "infer_schema" in kw:
        del kw["infer_schema"]
    header: Any = False
    if "header" in kw:
        header = kw["header"]
        del kw["header"]
    if str(header) in ["True", "0"]:
        pdf = _safe_load_csv(p, **{"index_col": False, "header": 0, **kw})
        if columns is None:
            return pdf, None
        if isinstance(columns, list):  # column names
            return pdf[columns], None
        schema = Schema(columns)
        return pdf[schema.names], schema
    if header is None or str(header) == "False":
        if columns is None:
            raise ValueError("columns must be set if without header")
        if isinstance(columns, list):  # column names
            pdf = _safe_load_csv(
                p,
                **{"index_col": False, "header": None, "names": columns, **kw},
            )
            return pdf, None
        schema = Schema(columns)
        pdf = _safe_load_csv(
            p,
            **{"index_col": False, "header": None, "names": schema.names, **kw},
        )
        return pdf, schema
    else:
        raise NotImplementedError(f"{header} is not supported")


def _save_json(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    with p.open("w") as f:
        df.as_pandas().to_json(f, **{"orient": "records", "lines": True, **kwargs})


def _safe_load_json(p: FileParser, **kwargs: Any) -> pd.DataFrame:
    kw = {"orient": "records", "lines": True, **kwargs}

    def load_dir() -> pd.DataFrame:
        dfs: List[pd.DataFrame] = []
        for _p in p.join("*.json").find_all():  # type: ignore
            with _p.open("r") as f:
                dfs.append(pd.read_json(f, **kw))
        return pd.concat(dfs)

    try:
        with p.open("r") as f:
            return pd.read_json(f, **kw)
    except (IsADirectoryError, PermissionError):
        return load_dir()


def _load_json(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    pdf = _safe_load_json(p, **kwargs).reset_index(drop=True)
    if columns is None:
        return pdf, None
    if isinstance(columns, list):  # column names
        return pdf[columns], None
    schema = Schema(columns)
    return pdf[schema.names], schema


_FORMAT_MAP: Dict[str, str] = {
    ".csv": "csv",
    ".csv.gz": "csv",
    ".parquet": "parquet",
    ".json": "json",
    ".json.gz": "json",
}

_FORMAT_LOAD: Dict[str, Callable[..., Tuple[pd.DataFrame, Any]]] = {
    "csv": _load_csv,
    "parquet": _load_parquet,
    "json": _load_json,
}

_FORMAT_SAVE: Dict[str, Callable] = {
    "csv": _save_csv,
    "parquet": _save_parquet,
    "json": _save_json,
}



from typing import Any, Optional

import pyarrow as pa
from fugue.column.expressions import (
    ColumnExpr,
    _FuncExpr,
    _to_col,
    function,
)
from triad import Schema


def coalesce(*args: Any) -> ColumnExpr:
    """SQL ``COALESCE`` function

    :param args: If a value is not :class:`~fugue.column.expressions.ColumnExpr`
      then it's converted to a literal column by
      :func:`~fugue.column.expressions.col`

    .. note::

        this function can infer neither type nor alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            f.coalesce(col("a"), col("b")+col("c"), 1)
    """
    return function("COALESCE", *[_to_col(x) for x in args])


def min(col: ColumnExpr) -> ColumnExpr:  # pylint: disable=redefined-builtin
    """SQL ``MIN`` function (aggregation)

    :param col: the column to find min

    .. note::

        * this function can infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            # assume col a has type double
            f.min(col("a"))  # CAST(MIN(a) AS double) AS a
            f.min(-col("a"))  # CAST(MIN(-a) AS double) AS a

            # neither type nor alias can be inferred in the following cases
            f.min(col("a")+1)
            f.min(col("a")+col("b"))

            # you can specify explicitly
            # CAST(MIN(a+b) AS int) AS x
            f.min(col("a")+col("b")).cast(int).alias("x")
    """
    assert isinstance(col, ColumnExpr)
    return _SameTypeUnaryAggFuncExpr("MIN", col)


def max(col: ColumnExpr) -> ColumnExpr:  # pylint: disable=redefined-builtin
    """SQL ``MAX`` function (aggregation)

    :param col: the column to find max

    .. note::

        * this function can infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            # assume col a has type double
            f.max(col("a"))  # CAST(MAX(a) AS double) AS a
            f.max(-col("a"))  # CAST(MAX(-a) AS double) AS a

            # neither type nor alias can be inferred in the following cases
            f.max(col("a")+1)
            f.max(col("a")+col("b"))

            # you can specify explicitly
            # CAST(MAX(a+b) AS int) AS x
            f.max(col("a")+col("b")).cast(int).alias("x")
    """
    assert isinstance(col, ColumnExpr)
    return _SameTypeUnaryAggFuncExpr("MAX", col)


def count(col: ColumnExpr) -> ColumnExpr:
    """SQL ``COUNT`` function (aggregation)

    :param col: the column to find count

    .. note::

        * this function cannot infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            f.count(all_cols())  # COUNT(*)
            f.count(col("a"))  # COUNT(a) AS a

            # you can specify explicitly
            # CAST(COUNT(a) AS double) AS a
            f.count(col("a")).cast(float)
    """
    assert isinstance(col, ColumnExpr)
    return _UnaryAggFuncExpr("COUNT", col)


def count_distinct(col: ColumnExpr) -> ColumnExpr:
    """SQL ``COUNT DISTINCT`` function (aggregation)

    :param col: the column to find distinct element count

    .. note::

        * this function cannot infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            f.count_distinct(all_cols())  # COUNT(DISTINCT *)
            f.count_distinct(col("a"))  # COUNT(DISTINCT a) AS a

            # you can specify explicitly
            # CAST(COUNT(DISTINCT a) AS double) AS a
            f.count_distinct(col("a")).cast(float)
    """
    assert isinstance(col, ColumnExpr)
    return _UnaryAggFuncExpr("COUNT", col, arg_distinct=True)


def avg(col: ColumnExpr) -> ColumnExpr:
    """SQL ``AVG`` function (aggregation)

    :param col: the column to find average

    .. note::

        * this function cannot infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            f.avg(col("a"))  # AVG(a) AS a

            # you can specify explicitly
            # CAST(AVG(a) AS double) AS a
            f.avg(col("a")).cast(float)
    """
    assert isinstance(col, ColumnExpr)
    return _UnaryAggFuncExpr("AVG", col)


def sum(col: ColumnExpr) -> ColumnExpr:  # pylint: disable=redefined-builtin
    """SQL ``SUM`` function (aggregation)

    :param col: the column to find sum

    .. note::

        * this function cannot infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            f.sum(col("a"))  # SUM(a) AS a

            # you can specify explicitly
            # CAST(SUM(a) AS double) AS a
            f.sum(col("a")).cast(float)
    """
    assert isinstance(col, ColumnExpr)
    return _UnaryAggFuncExpr("SUM", col)


def first(col: ColumnExpr) -> ColumnExpr:
    """SQL ``FIRST`` function (aggregation)

    :param col: the column to find first

    .. note::

        * this function can infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            # assume col a has type double
            f.first(col("a"))  # CAST(FIRST(a) AS double) AS a
            f.first(-col("a"))  # CAST(FIRST(-a) AS double) AS a

            # neither type nor alias can be inferred in the following cases
            f.first(col("a")+1)
            f.first(col("a")+col("b"))

            # you can specify explicitly
            # CAST(FIRST(a+b) AS int) AS x
            f.first(col("a")+col("b")).cast(int).alias("x")
    """
    assert isinstance(col, ColumnExpr)
    return _SameTypeUnaryAggFuncExpr("FIRST", col)


def last(col: ColumnExpr) -> ColumnExpr:
    """SQL ``LAST`` function (aggregation)

    :param col: the column to find last

    .. note::

        * this function can infer type from ``col`` type
        * this function can infer alias from ``col``'s inferred alias

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            # assume col a has type double
            f.last(col("a"))  # CAST(LAST(a) AS double) AS a
            f.last(-col("a"))  # CAST(LAST(-a) AS double) AS a

            # neither type nor alias can be inferred in the following cases
            f.last(col("a")+1)
            f.last(col("a")+col("b"))

            # you can specify explicitly
            # CAST(LAST(a+b) AS int) AS x
            f.last(col("a")+col("b")).cast(int).alias("x")
    """
    assert isinstance(col, ColumnExpr)
    return _SameTypeUnaryAggFuncExpr("LAST", col)


def is_agg(column: Any) -> bool:
    """Check if a column contains aggregation operation

    :param col: the column to check
    :return: whether the column is :class:`~fugue.column.expressions.ColumnExpr`
      and contains aggregation operations

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column.functions as f

            assert not f.is_agg(1)
            assert not f.is_agg(col("a"))
            assert not f.is_agg(col("a")+lit(1))

            assert f.is_agg(f.max(col("a")))
            assert f.is_agg(-f.max(col("a")))
            assert f.is_agg(f.max(col("a")+1))
            assert f.is_agg(f.max(col("a"))+f.min(col("a"))))
    """
    if isinstance(column, _UnaryAggFuncExpr):
        return True
    if isinstance(column, _FuncExpr):
        return any(is_agg(x) for x in column.args) or any(
            is_agg(x) for x in column.kwargs.values()
        )
    return False


class _UnaryAggFuncExpr(_FuncExpr):
    def __init__(self, func: str, col: ColumnExpr, arg_distinct: bool = False):
        super().__init__(func, col, arg_distinct=arg_distinct)

    def infer_alias(self) -> ColumnExpr:
        return (
            self
            if self.output_name != ""
            else self.alias(self.args[0].infer_alias().output_name)
        )

    def _copy(self) -> _FuncExpr:
        return _UnaryAggFuncExpr(self.func, *self.args, **self.kwargs)


class _SameTypeUnaryAggFuncExpr(_UnaryAggFuncExpr):
    def _copy(self) -> _FuncExpr:
        return _SameTypeUnaryAggFuncExpr(self.func, *self.args, **self.kwargs)

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        return self.as_type or self.args[0].infer_type(schema)



import pickle
from typing import Any, Iterable, List, Tuple, Optional


import pandas as pd
import pyarrow as pa
import pyspark
import pyspark.sql as ps
import pyspark.sql.types as pt
from packaging import version
from pyarrow.types import is_list, is_struct, is_timestamp
from pyspark.sql.pandas.types import (
    from_arrow_schema,
    from_arrow_type,
    to_arrow_schema,
    to_arrow_type,
)
from triad.collections import Schema
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.pyarrow import TRIAD_DEFAULT_TIMESTAMP, cast_pa_table
from triad.utils.schema import quote_name

import fugue.api as fa
from fugue import DataFrame

from .misc import is_spark_dataframe

try:
    from pyspark.sql.types import TimestampNTZType  # pylint: disable-all
except ImportError:  # pragma: no cover
    # pyspark < 3.2
    from pyspark.sql.types import TimestampType as TimestampNTZType

_PYSPARK_ARROW_FRIENDLY = version.parse(pyspark.__version__) >= version.parse("3.3")


def pandas_udf_can_accept(schema: Schema, is_input: bool) -> bool:
    try:
        # pyspark's own from_arrow_schema to_arrow_schema
        # can validate if a type can be supported by pandas udf
        if not is_input and any(pa.types.is_struct(t) for t in schema.types):
            return False
        to_arrow_schema(from_arrow_schema(schema.pa_schema))
        return True
    except Exception:  # pragma: no cover
        return False


def to_spark_schema(obj: Any) -> pt.StructType:
    assert_arg_not_none(obj, "schema")
    if isinstance(obj, pt.StructType):
        return obj
    if is_spark_dataframe(obj):
        return obj.schema
    return _from_arrow_schema(Schema(obj).pa_schema)


def to_schema(obj: Any) -> Schema:
    assert_arg_not_none(obj, "obj")
    if isinstance(obj, pt.StructType):
        return Schema(_to_arrow_schema(obj))
    if is_spark_dataframe(obj):
        return to_schema(obj.schema)
    return Schema(obj)


def to_cast_expression(
    schema1: Any, schema2: Any, allow_name_mismatch: bool
) -> Tuple[bool, List[str]]:
    schema1 = to_spark_schema(schema1)
    schema2 = to_spark_schema(schema2)
    assert_or_throw(
        len(schema1) == len(schema2),
        lambda: ValueError(f"schema mismatch: {schema1}, {schema2}"),
    )
    expr: List[str] = []
    has_cast = False
    for i in range(len(schema1)):
        name_match = schema1[i].name == schema2[i].name
        assert_or_throw(
            name_match or allow_name_mismatch,
            lambda: ValueError(f"schema name mismatch: {schema1}, {schema2}"),
        )
        n1, n2 = quote_name(schema1[i].name, quote="`"), quote_name(
            schema2[i].name, quote="`"
        )
        if schema1[i].dataType != schema2[i].dataType:
            type2 = schema2[i].dataType.simpleString()
            if isinstance(schema1[i].dataType, pt.FractionalType) and isinstance(
                schema2[i].dataType, (pt.StringType, pt.IntegralType)
            ):
                expr.append(
                    f"CAST(IF(isnan({n1}) OR {n1} IS NULL"
                    f", NULL, {n1})"
                    f" AS {type2}) {n2}"
                )
            else:
                expr.append(f"CAST({n1} AS {type2}) {n2}")
            has_cast = True
        else:
            if schema1[i].name != schema2[i].name:
                expr.append(f"{n1} AS {n2}")
                has_cast = True
            else:
                expr.append(n1)
    return has_cast, expr


def to_select_expression(schema_from: Any, schema_to: Any) -> List[str]:
    schema1 = to_spark_schema(schema_from)
    if isinstance(schema_to, List):
        return [schema1[n].name for n in schema_to]
    schema2 = to_spark_schema(schema_to)
    sub = pt.StructType([schema1[x.name] for x in schema2.fields])
    _, expr = to_cast_expression(sub, schema2, allow_name_mismatch=False)
    return expr


def to_type_safe_input(rows: Iterable[ps.Row], schema: Schema) -> Iterable[List[Any]]:
    struct_idx = [p for p, t in enumerate(schema.types) if pa.types.is_struct(t)]
    complex_list_idx = [
        p
        for p, t in enumerate(schema.types)
        if pa.types.is_list(t) and pa.types.is_nested(t.value_type)
    ]
    if len(struct_idx) == 0 and len(complex_list_idx) == 0:
        for row in rows:
            yield list(row)
    elif len(complex_list_idx) == 0:
        for row in rows:
            r = list(row)
            for i in struct_idx:
                if r[i] is not None:
                    r[i] = r[i].asDict(recursive=True)
            yield r
    else:  # pragma: no cover
        for row in rows:
            data = row.asDict(recursive=True)
            r = [data[n] for n in schema.names]
            yield r


def to_spark_df(session: ps.SparkSession, df: Any, schema: Any = None) -> ps.DataFrame:
    if schema is not None and not isinstance(schema, pt.StructType):
        schema = to_spark_schema(schema)
    if isinstance(df, pd.DataFrame):
        if pd.__version__ >= "2" and session.version < "3.4":  # pragma: no cover
            # pyspark < 3.4 does not support pandas 2 when doing
            # createDataFrame, see this issue:
            # https://stackoverflow.com/a/75926954/12309438
            # this is a workaround with the cost of memory and speed.
            if schema is None:
                schema = to_spark_schema(fa.get_schema(df))
            df = fa.as_fugue_df(df).as_array(type_safe=True)
        return pd_to_spark_df(session, df, schema=schema)
    if isinstance(df, DataFrame):
        if pd.__version__ >= "2" and session.version < "3.4":  # pragma: no cover
            if schema is None:
                schema = to_spark_schema(df.schema)
            return session.createDataFrame(df.as_array(type_safe=True), schema=schema)
        return pd_to_spark_df(session, df.as_pandas(), schema=schema)
    else:
        return session.createDataFrame(df, schema=schema)


def pd_to_spark_df(
    session: ps.SparkSession, df: pd.DataFrame, schema: Any = None
) -> ps.DataFrame:
    if _PYSPARK_ARROW_FRIENDLY:
        return session.createDataFrame(df, schema=schema)
    else:  # pragma: no cover
        # Not efficient for pyspark<3.3
        return session.createDataFrame(df.astype(object), schema=schema)


def to_pandas(df: ps.DataFrame) -> pd.DataFrame:
    if version.parse(pd.__version__) < version.parse("2.0.0") or not any(
        isinstance(x.dataType, (pt.TimestampType, TimestampNTZType))
        for x in df.schema.fields
    ):
        return df.toPandas()
    else:  # pragma: no cover

        def serialize(dfs):
            for df in dfs:
                data = pickle.dumps(df)
                yield pd.DataFrame([[data]], columns=["data"])

        sdf = df.mapInPandas(serialize, schema="data binary")
        return pd.concat(pickle.loads(x.data) for x in sdf.collect())


def to_arrow(df: ps.DataFrame) -> pa.Table:
    schema = to_schema(df.schema)
    destruct: Optional[bool] = None
    try:
        jconf = df.sparkSession._jconf
        if jconf.arrowPySparkEnabled() and pandas_udf_can_accept(
            schema, is_input=False
        ):
            destruct = jconf.arrowPySparkSelfDestructEnabled()
    except Exception:  # pragma: no cover
        # older spark does not have this config
        pass
    if destruct is not None and hasattr(df, "_collect_as_arrow"):
        batches = df._collect_as_arrow(split_batches=destruct)
        if len(batches) == 0:
            return schema.create_empty_arrow_table()
        table = pa.Table.from_batches(batches)
        del batches
        return cast_pa_table(table, schema.pa_schema)
    else:  # pragma: no cover
        # df.toPandas has bugs on nested types
        return pa.Table.from_pylist(df.collect(), schema=schema.pa_schema)


# TODO: the following function always set nullable to true,
# but should we use field.nullable?
def _to_arrow_type(dt: pt.DataType) -> pa.DataType:
    if isinstance(dt, pt.TimestampType):
        return TRIAD_DEFAULT_TIMESTAMP
    if isinstance(dt, pt.StructType):
        fields = [
            pa.field(
                # field.name, _to_arrow_type(field.dataType), nullable=field.nullable
                field.name,
                _to_arrow_type(field.dataType),
                nullable=True,
            )
            for field in dt
        ]
        return pa.struct(fields)
    if isinstance(dt, pt.ArrayType):
        return pa.list_(_to_arrow_type(dt.elementType))
    if isinstance(dt, pt.MapType):
        return pa.map_(_to_arrow_type(dt.keyType), _to_arrow_type(dt.valueType))
    return to_arrow_type(dt)


def _to_arrow_schema(schema: pt.StructType) -> pa.Schema:
    fields = [
        # pa.field(field.name, _to_arrow_type(field.dataType), nullable=field.nullable)
        pa.field(field.name, _to_arrow_type(field.dataType), nullable=True)
        for field in schema
    ]
    return pa.schema(fields)


def _from_arrow_type(dt: pa.DataType) -> pt.DataType:
    if is_struct(dt):
        return pt.StructType(
            [
                pt.StructField(
                    # field.name, _from_arrow_type(field.type), nullable=field.nullable
                    field.name,
                    _from_arrow_type(field.type),
                    nullable=True,
                )
                for field in dt
            ]
        )
    elif is_list(dt):
        if is_timestamp(dt.value_type):
            raise TypeError(  # pragma: no cover
                "Spark: unsupported type in conversion from Arrow: " + str(dt)
            )
        return pt.ArrayType(_from_arrow_type(dt.value_type))
    return from_arrow_type(dt)


def _from_arrow_schema(schema: pa.Schema) -> pt.StructType:
    return pt.StructType(
        [
            pt.StructField(
                # field.name, _from_arrow_type(field.type), nullable=field.nullable
                field.name,
                _from_arrow_type(field.type),
                nullable=True,
            )
            for field in schema
        ]
    )



from typing import Any, List, Type, no_type_check

from triad.collections import ParamDict
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import to_type

from fugue.collections.partition import PartitionCursor
from fugue.column import ColumnExpr
from fugue.column import SelectColumns as ColumnsSelect
from fugue.dataframe import ArrayDataFrame, DataFrame, DataFrames, LocalDataFrame
from fugue.exceptions import FugueWorkflowError
from fugue.execution import make_sql_engine
from fugue.execution.execution_engine import (
    _FUGUE_SERIALIZED_BLOB_SCHEMA,
    _generate_comap_empty_dfs,
)
from fugue.extensions.processor import Processor
from fugue.extensions.transformer import CoTransformer, Transformer, _to_transformer
from fugue.rpc import EmptyRPCHandler, to_rpc_handler


class RunTransformer(Processor):
    @no_type_check
    def process(self, dfs: DataFrames) -> DataFrame:
        df = dfs[0]
        tf = _to_transformer(
            self.params.get_or_none("transformer", object),
            self.params.get_or_none("schema", object),
        )
        tf._workflow_conf = self.execution_engine.conf
        tf._params = self.params.get("params", ParamDict())  # type: ignore
        tf._partition_spec = self.partition_spec
        rpc_handler = to_rpc_handler(self.params.get_or_throw("rpc_handler", object))
        if not isinstance(rpc_handler, EmptyRPCHandler):
            tf._rpc_client = self.rpc_server.make_client(rpc_handler)
        ie = self.params.get("ignore_errors", [])
        self._ignore_errors = [to_type(x, Exception) for x in ie]
        tf.validate_on_runtime(df)
        if isinstance(tf, Transformer):
            return self.transform(df, tf)
        else:
            return self.cotransform(df, tf)

    def transform(self, df: DataFrame, tf: Transformer) -> DataFrame:
        tf._key_schema = self.partition_spec.get_key_schema(df.schema)  # type: ignore
        tf._output_schema = Schema(tf.get_output_schema(df))  # type: ignore
        tr = _TransformerRunner(df, tf, self._ignore_errors)  # type: ignore
        return self.execution_engine.map_engine.map_dataframe(
            df=df,
            map_func=tr.run,
            output_schema=tf.output_schema,  # type: ignore
            partition_spec=tf.partition_spec,
            on_init=tr.on_init,
            map_func_format_hint=tf.get_format_hint(),
        )

    @no_type_check
    def cotransform(self, df: DataFrame, tf: CoTransformer) -> DataFrame:
        assert_or_throw(
            df.metadata.get("serialized", False), "must use serialized dataframe"
        )
        tf._key_schema = df.schema - _FUGUE_SERIALIZED_BLOB_SCHEMA
        # TODO: currently, get_output_schema only gets empty dataframes
        empty_dfs = _generate_comap_empty_dfs(
            df.metadata["schemas"], df.metadata.get("serialized_has_name", False)
        )
        tf._output_schema = Schema(tf.get_output_schema(empty_dfs))
        tr = _CoTransformerRunner(df, tf, self._ignore_errors)
        return self.execution_engine.comap(
            df=df,
            map_func=tr.run,
            output_schema=tf.output_schema,
            partition_spec=tf.partition_spec,
            on_init=tr.on_init,
        )


class RunJoin(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        if len(dfs) == 1:
            return dfs[0]
        how = self.params.get_or_throw("how", str)
        on = self.params.get("on", [])
        df = dfs[0]
        for i in range(1, len(dfs)):
            df = self.execution_engine.join(df, dfs[i], how=how, on=on)
        return df


class RunSetOperation(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        if len(dfs) == 1:
            return dfs[0]
        how = self.params.get_or_throw("how", str)
        func: Any = {
            "union": self.execution_engine.union,
            "subtract": self.execution_engine.subtract,
            "intersect": self.execution_engine.intersect,
        }[how]
        distinct = self.params.get("distinct", True)
        df = dfs[0]
        for i in range(1, len(dfs)):
            df = func(df, dfs[i], distinct=distinct)
        return df


class Distinct(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        return self.execution_engine.distinct(dfs[0])


class Dropna(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        how = self.params.get("how", "any")
        assert_or_throw(
            how in ["any", "all"],
            FugueWorkflowError("how' needs to be either 'any' or 'all'"),
        )
        thresh = self.params.get_or_none("thresh", int)
        subset = self.params.get_or_none("subset", list)
        return self.execution_engine.dropna(
            dfs[0], how=how, thresh=thresh, subset=subset
        )


class Fillna(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        value = self.params.get_or_none("value", object)
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            FugueWorkflowError("fillna value cannot be None or list"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                FugueWorkflowError(
                    "fillna dict can't contain None and must have len > 1"
                ),
            )
        subset = self.params.get_or_none("subset", list)
        return self.execution_engine.fillna(dfs[0], value=value, subset=subset)


class RunSQLSelect(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        statement = self.params.get_or_throw("statement", object)
        engine = self.params.get_or_none("sql_engine", object)
        engine_params = self.params.get("sql_engine_params", ParamDict())
        sql_engine = make_sql_engine(engine, self.execution_engine, **engine_params)
        return sql_engine.select(dfs, statement)


class Zip(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        how = self.params.get("how", "inner")
        partition_spec = self.partition_spec
        # TODO: this should also search on workflow conf
        temp_path = self.params.get_or_none("temp_path", str)
        to_file_threshold = self.params.get_or_none("to_file_threshold", object)
        return self.execution_engine.zip(
            dfs,
            how=how,
            partition_spec=partition_spec,
            temp_path=temp_path,
            to_file_threshold=to_file_threshold,
        )


class Select(Processor):
    def validate_on_compile(self):
        sc = self.params.get_or_throw("columns", ColumnsSelect)
        sc.assert_all_with_names()

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", ColumnsSelect)
        where = None if "where" not in self.params else self.params["where"]
        having = None if "having" not in self.params else self.params["having"]
        return self.execution_engine.select(
            df=dfs[0], cols=columns, where=where, having=having
        )


class Filter(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("condition", ColumnExpr)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        condition = self.params.get_or_throw("condition", ColumnExpr)
        return self.execution_engine.filter(df=dfs[0], condition=condition)


class Assign(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("columns", list)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", list)
        return self.execution_engine.assign(df=dfs[0], columns=columns)


class Aggregate(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("columns", list)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", list)
        return self.execution_engine.aggregate(
            df=dfs[0], partition_spec=self.partition_spec, agg_cols=columns
        )


class Rename(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("columns", dict)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", dict)
        return dfs[0].rename(columns)


class AlterColumns(Processor):
    def validate_on_compile(self):
        Schema(self.params.get_or_throw("columns", object))

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", object)
        return dfs[0].alter_columns(columns)


class DropColumns(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("columns", list)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        if_exists = self.params.get("if_exists", False)
        columns = self.params.get_or_throw("columns", list)
        if if_exists:
            columns = set(columns).intersection(dfs[0].schema.keys())
        return dfs[0].drop(list(columns))


class SelectColumns(Processor):
    def validate_on_compile(self):
        self.params.get_or_throw("columns", list)

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        columns = self.params.get_or_throw("columns", list)
        return dfs[0][columns]


class Sample(Processor):
    def validate_on_compile(self):
        n = self.params.get_or_none("n", int)
        frac = self.params.get_or_none("frac", float)
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )

    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        n = self.params.get_or_none("n", int)
        frac = self.params.get_or_none("frac", float)
        replace = self.params.get("replace", False)
        seed = self.params.get_or_none("seed", int)
        return self.execution_engine.sample(
            dfs[0], n=n, frac=frac, replace=replace, seed=seed
        )


class Take(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        # All _get_or operations convert float to int
        n = self.params.get_or_none("n", int)
        presort = self.params.get_or_none("presort", str)
        na_position = self.params.get("na_position", "last")
        partition_spec = self.partition_spec
        return self.execution_engine.take(
            dfs[0],
            n,
            presort=presort,
            na_position=na_position,
            partition_spec=partition_spec,
        )


class SaveAndUse(Processor):
    def process(self, dfs: DataFrames) -> DataFrame:
        assert_or_throw(len(dfs) == 1, FugueWorkflowError("not single input"))
        kwargs = self.params.get("params", dict())
        path = self.params.get_or_throw("path", str)
        format_hint = self.params.get("fmt", "")
        mode = self.params.get("mode", "overwrite")
        partition_spec = self.partition_spec
        force_single = self.params.get("single", False)

        self.execution_engine.save_df(
            df=dfs[0],
            path=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=partition_spec,
            force_single=force_single,
            **kwargs
        )
        return self.execution_engine.load_df(path=path, format_hint=format_hint)


class _TransformerRunner(object):
    def __init__(
        self, df: DataFrame, transformer: Transformer, ignore_errors: List[type]
    ):
        self.schema = df.schema
        self.transformer = transformer
        self.ignore_errors = tuple(ignore_errors)

    def run(self, cursor: PartitionCursor, df: LocalDataFrame) -> LocalDataFrame:
        self.transformer._cursor = cursor  # type: ignore
        if len(self.ignore_errors) == 0:
            return self.transformer.transform(df)
        else:
            try:
                return self.transformer.transform(df).as_local_bounded()
            except self.ignore_errors:  # type: ignore  # pylint: disable=E0712
                return ArrayDataFrame([], self.transformer.output_schema)

    def on_init(self, partition_no: int, df: DataFrame) -> None:
        s = self.transformer.partition_spec
        self.transformer._cursor = s.get_cursor(  # type: ignore
            self.schema, partition_no
        )
        self.transformer.on_init(df)


class _CoTransformerRunner(object):
    def __init__(
        self,
        df: DataFrame,
        transformer: CoTransformer,
        ignore_errors: List[Type[Exception]],
    ):
        self.schema = df.schema
        self.transformer = transformer
        self.ignore_errors = tuple(ignore_errors)

    def run(self, cursor: PartitionCursor, dfs: DataFrames) -> LocalDataFrame:
        self.transformer._cursor = cursor  # type: ignore
        if len(self.ignore_errors) == 0:
            return self.transformer.transform(dfs)

        else:
            try:
                return self.transformer.transform(dfs).as_local_bounded()
            except self.ignore_errors:  # type: ignore  # pylint: disable=E0712
                return ArrayDataFrame([], self.transformer.output_schema)

    def on_init(self, partition_no: int, dfs: DataFrames) -> None:
        s = self.transformer.partition_spec
        self.transformer._cursor = s.get_cursor(  # type: ignore
            self.schema, partition_no
        )
        self.transformer.on_init(dfs)



from typing import Any, Callable, Dict

from triad import Schema
from triad.utils.assertion import assert_or_throw

from fugue._utils.interfaceless import parse_comment_annotation
from fugue.collections.partition import PartitionSpec, parse_presort_exp
from fugue.exceptions import (
    FugueWorkflowCompileValidationError,
    FugueWorkflowRuntimeValidationError,
)


def is_namespace_extension(obj: Any) -> bool:
    return isinstance(obj, tuple) and len(obj) == 2 and isinstance(obj[0], str)


def load_namespace_extensions(obj: Any) -> None:
    if is_namespace_extension(obj):
        from fugue_contrib import load_namespace

        load_namespace(obj[0])


def namespace_candidate(
    namespace: str, matcher: Callable[..., bool]
) -> Callable[..., bool]:
    def _matcher(obj: Any, *args: Any, **kwargs: Any) -> bool:
        if is_namespace_extension(obj) and obj[0] == namespace:
            return matcher(obj[1], *args, **kwargs)
        return False

    return _matcher


def parse_validation_rules_from_comment(func: Callable) -> Dict[str, Any]:
    res: Dict[str, Any] = {}
    for key in [
        "partitionby_has",
        "partitionby_is",
        "presort_has",
        "presort_is",
        "input_has",
        "input_is",
    ]:
        v = parse_comment_annotation(func, key)
        if v is None:
            continue
        assert_or_throw(v != "", lambda: SyntaxError(f"{key} can't be empty"))
        res[key] = v
    return to_validation_rules(res)


def to_validation_rules(data: Dict[str, Any]) -> Dict[str, Any]:
    res: Dict[str, Any] = {}
    for k, v in data.items():
        if k in ["partitionby_has", "partitionby_is"]:
            if isinstance(v, str):
                v = [x.strip() for x in v.split(",")]
            res[k] = PartitionSpec(by=v).partition_by
        elif k in ["presort_has", "presort_is"]:
            res[k] = list(parse_presort_exp(v).items())
        elif k in ["input_has"]:
            if isinstance(v, str):
                res[k] = v.replace(" ", "").split(",")
            else:
                assert_or_throw(
                    isinstance(v, list),
                    lambda: SyntaxError(f"{v} is neither a string or a list"),
                )
                res[k] = [x.replace(" ", "") for x in v]
        elif k in ["input_is"]:
            try:
                res[k] = str(Schema(v))
            except SyntaxError:
                raise SyntaxError(  # pylint: disable=W0707
                    f"for input_is, the input must be a schema expression {v}"
                )
        else:
            raise NotImplementedError(k)
    return res


def validate_partition_spec(spec: PartitionSpec, rules: Dict[str, Any]) -> None:
    for k, v in rules.items():
        if k in ["partitionby_has", "partitionby_is"]:
            for x in v:
                assert_or_throw(
                    x in spec.partition_by,
                    lambda: FugueWorkflowCompileValidationError(
                        f"required partition key {x} is not in {spec}"
                    ),
                )
            if k == "partitionby_is":
                assert_or_throw(
                    len(v) == len(spec.partition_by),
                    lambda: FugueWorkflowCompileValidationError(
                        f"{v} does not match {spec}"
                    ),
                )
        if k in ["presort_has", "presort_is"]:
            expected = spec.presort
            for pk, pv in v:
                o = "ASC" if pv else "DESC"
                assert_or_throw(
                    pk in expected,
                    lambda: FugueWorkflowCompileValidationError(
                        f"required presort key {pk} is not in presort of {spec}"
                    ),
                )
                assert_or_throw(
                    pv == expected[pk],
                    lambda: FugueWorkflowCompileValidationError(
                        f"({pk},{o}) order does't match presort of {spec}"
                    ),
                )
            if k == "presort_is":
                assert_or_throw(
                    len(v) == len(expected),
                    lambda: FugueWorkflowCompileValidationError(
                        f"{v} does not match {spec}"
                    ),
                )
                assert_or_throw(
                    v == list(expected.items()),
                    lambda: FugueWorkflowCompileValidationError(
                        f"{v} order does not match {spec}"
                    ),
                )


def validate_input_schema(schema: Schema, rules: Dict[str, Any]) -> None:
    for k, v in rules.items():
        if k == "input_has":
            for x in v:
                assert_or_throw(
                    x in schema,
                    lambda: FugueWorkflowRuntimeValidationError(
                        f"required column {x} is not in {schema}"
                    ),
                )
        if k == "input_is":
            assert_or_throw(
                schema == v,
                lambda: FugueWorkflowRuntimeValidationError(
                    f"{v} does not match {schema}"
                ),
            )



import json
from typing import Any, Dict, List, Tuple

from triad.collections.dict import IndexedOrderedDict, ParamDict
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw as aot
from triad.utils.convert import to_size
from triad.utils.hash import to_uuid
from triad.utils.pyarrow import SchemaedDataPartitioner
from triad.utils.schema import safe_split_out_of_quote, unquote_name


def parse_presort_exp(presort: Any) -> IndexedOrderedDict[str, bool]:  # noqa [C901]
    """Returns ordered column sorting direction where ascending order
    would return as true, and descending as false.

    :param presort: string that contains column and sorting direction or
        list of tuple that contains column and boolean sorting direction
    :type presort: Any

    :return: column and boolean sorting direction of column, order matters.
    :rtype: IndexedOrderedDict[str, bool]

    .. admonition:: Examples

        >>> parse_presort_exp("b desc, c asc")
        >>> parse_presort_exp([("b", True), ("c", False))])
        both return IndexedOrderedDict([("b", True), ("c", False))])
    """

    if isinstance(presort, IndexedOrderedDict):
        return presort

    presort_list: List[Tuple[str, bool]] = []
    res: IndexedOrderedDict[str, bool] = IndexedOrderedDict()
    if presort is None:
        return res

    elif isinstance(presort, str):
        presort = presort.strip()
        if presort == "":
            return res
        for p in safe_split_out_of_quote(presort, ","):
            pp = safe_split_out_of_quote(p.strip(), " ", max_split=1)
            key = unquote_name(pp[0].strip())
            if len(pp) == 1:
                presort_list.append((key, True))
            elif len(pp) == 2:
                if pp[1].strip().lower() == "asc":
                    presort_list.append((key, True))
                elif pp[1].strip().lower() == "desc":
                    presort_list.append((key, False))
                else:
                    raise SyntaxError(f"Invalid expression {presort}")
            else:  # pragma: no cover
                # impossible
                raise SyntaxError(f"Invalid expression {presort}")

    elif isinstance(presort, list):
        for p in presort:
            if isinstance(p, str):
                presort_list.append((p, True))
            else:
                aot(len(p) == 2, SyntaxError(f"Invalid expression {presort}"))
                aot(
                    isinstance(p, tuple)
                    & (isinstance(p[0], str) & (isinstance(p[1], bool))),
                    SyntaxError(f"Invalid expression {presort}"),
                )
                presort_list.append((p[0], p[1]))

    for key, value in presort_list:
        if key in res:
            raise SyntaxError(f"Invalid expression {presort} duplicated key {key}")
        res[key] = value
    return res


class PartitionSpec(object):
    """Fugue Partition Specification.

    .. admonition:: Examples

        >>> PartitionSepc(num=4)
        >>> PartitionSepc(4)  # == PartitionSepc(num=4)
        >>> PartitionSepc(num="ROWCOUNT/4 + 3")  # It can be an expression
        >>> PartitionSepc(by=["a","b"])
        >>> PartitionSepc(["a","b"])  # == PartitionSepc(by=["a","b"])
        >>> PartitionSpec(by=["a"], presort="b DESC, c ASC")
        >>> PartitionSpec(algo="even", num=4)
        >>> p = PartitionSpec(num=4, by=["a"])
        >>> p_override = PartitionSpec(p, by=["a","b"], algo="even")
        >>> PartitionSpec(by="a")  # == PartitionSpec(by=["a"])
        >>> PartitionSpec("a")  # == PartitionSpec(by=["a"])
        >>> PartitionSpec("per_row")  # == PartitionSpec(num="ROWCOUNT", algo="even")

    It's important to understand this concept, please read |PartitionTutorial|

    Partition consists for these specs:

    * **algo**: can be one of ``default``, ``hash``, ``rand``, ``even`` or ``coarse``
    * **num** or **num_partitions**: number of physical partitions, it can be an
      expression or integer numbers, e.g ``(ROWCOUNT+4) / 3``
    * **by** or **partition_by**: keys to partition on
    * **presort**: keys to sort other than partition keys. E.g. ``a``
      and ``a asc`` means presort by column a ascendingly, ``a,b desc``
      means presort by a ascendingly and then by b descendingly.
    * row_limit and size_limit are to be deprecated
    """

    def __init__(self, *args: Any, **kwargs: Any):  # noqa: C901
        p = ParamDict()
        if len(args) == 1 and len(kwargs) == 0:
            if isinstance(args[0], str):
                if args[0].lower() == "per_row":
                    p["algo"] = "even"
                    p["num_partitions"] = "ROWCOUNT"
                elif not args[0].startswith("{"):
                    p["partition_by"] = [args[0]]
            elif isinstance(args[0], int):
                p["num_partitions"] = str(args[0])
            elif isinstance(args[0], (list, tuple)):
                p["partition_by"] = args[0]
        if len(p) == 0:  # the first condition had no match
            for a in args:
                if a is None:
                    continue
                elif isinstance(a, PartitionSpec):
                    self._update_dict(p, a.jsondict)
                elif isinstance(a, Dict):
                    self._update_dict(p, a)
                elif isinstance(a, str):
                    self._update_dict(p, json.loads(a))
                else:
                    raise TypeError(f"{a} is not supported")
            self._update_dict(p, kwargs)
        self._num_partitions = p.get("num_partitions", "0")
        self._algo = p.get("algo", "").lower()
        if "partition_by" not in p:
            self._partition_by: List[str] = []
        elif isinstance(p["partition_by"], str):
            self._partition_by = [p["partition_by"]]
        elif isinstance(p["partition_by"], (list, tuple)):
            self._partition_by = list(p["partition_by"])
        else:
            raise SyntaxError(p["partition_by"])
        aot(
            len(self._partition_by) == len(set(self._partition_by)),
            SyntaxError(f"{self._partition_by} has duplicated keys"),
        )
        self._presort = parse_presort_exp(p.get_or_none("presort", object))
        if any(x in self._presort for x in self._partition_by):
            raise SyntaxError(
                "partition by overlap with presort: "
                + f"{self._partition_by}, {self._presort}"
            )
        # TODO: currently, size limit not in use
        self._size_limit = to_size(p.get("size_limit", "0"))
        self._row_limit = p.get("row_limit", 0)

    def __repr__(self) -> str:
        return (
            f"PartitionSpec(num='{self._num_partitions}', "
            + f"by={self._partition_by}, presort='{self.presort_expr}')"
        )

    def __eq__(self, other: Any) -> bool:
        if other is self:
            return True
        if not isinstance(other, PartitionSpec):
            other = PartitionSpec(other)
        return self.jsondict == other.jsondict

    @property
    def empty(self) -> bool:
        """Whether this spec didn't specify anything"""
        return (
            self._num_partitions == "0"
            and self._algo == ""
            and len(self._partition_by) == 0
            and len(self._presort) == 0
            and self._size_limit == 0
            and self._row_limit == 0
        )

    @property
    def num_partitions(self) -> str:
        """Number of partitions, it can be a string expression or int"""
        return self._num_partitions

    def get_num_partitions(self, **expr_map_funcs: Any) -> int:
        """Convert ``num_partitions`` expression to int number

        :param expr_map_funcs: lambda functions (no parameter) for keywords
        :return: integer value of the partitions

        .. admonition:: Examples

            >>> p = PartitionSpec(num="ROWCOUNT/2")
            >>> p.get_num_partitions(ROWCOUNT=lambda: df.count())
        """
        expr = self.num_partitions
        for k, v in expr_map_funcs.items():
            if k in expr:
                value = str(v())
                expr = expr.replace(k, value)
        return int(eval(expr))  # pylint: disable=W0123

    @property
    def algo(self) -> str:
        """Get algo of the spec, one of ``default``, ``hash``,
        ``rand`` ``even`` or ``coarse``
        """
        return self._algo if self._algo != "" else "default"

    @property
    def partition_by(self) -> List[str]:
        """Get partition keys of the spec"""
        return self._partition_by

    @property
    def presort(self) -> IndexedOrderedDict[str, bool]:
        """Get presort pairs of the spec

        .. admonition:: Examples

            >>> p = PartitionSpec(by=["a"],presort="b,c desc")
            >>> assert p.presort == {"b":True, "c":False}
        """
        return self._presort

    @property
    def presort_expr(self) -> str:
        """Get normalized presort expression

        .. admonition:: Examples

            >>> p = PartitionSpec(by=["a"],presort="b , c dESc")
            >>> assert p.presort_expr == "b ASC,c DESC"
        """
        return ",".join(
            k + " " + ("ASC" if v else "DESC") for k, v in self.presort.items()
        )

    @property
    def jsondict(self) -> ParamDict:
        """Get json serializeable dict of the spec"""
        return ParamDict(
            dict(
                num_partitions=self._num_partitions,
                algo=self._algo,
                partition_by=self._partition_by,
                presort=self.presort_expr,
                size_limit=self._size_limit,
                row_limit=self._row_limit,
            )
        )

    def __uuid__(self) -> str:
        """Get deterministic unique id of this object"""
        return to_uuid(self.jsondict)

    def get_sorts(
        self, schema: Schema, with_partition_keys: bool = True
    ) -> IndexedOrderedDict[str, bool]:
        """Get keys for sorting in a partition, it's the combination of partition
        keys plus the presort keys

        :param schema: the dataframe schema this partition spec to operate on
        :param with_partition_keys: whether to include partition keys
        :return: an ordered dictionary of key, order pairs

        .. admonition:: Examples

            >>> p = PartitionSpec(by=["a"],presort="b , c dESc")
            >>> schema = Schema("a:int,b:int,c:int,d:int"))
            >>> assert p.get_sorts(schema) == {"a":True, "b":True, "c": False}
        """
        d: IndexedOrderedDict[str, bool] = IndexedOrderedDict()
        if with_partition_keys:
            for p in self.partition_by:
                aot(p in schema, lambda: KeyError(f"{p} not in {schema}"))
                d[p] = True
        for p, v in self.presort.items():
            aot(p in schema, lambda: KeyError(f"{p} not in {schema}"))
            d[p] = v
        return d

    def get_key_schema(self, schema: Schema) -> Schema:
        """Get partition keys schema

        :param schema: the dataframe schema this partition spec to operate on
        :return: the sub-schema only containing partition keys
        """
        return schema.extract(self.partition_by)

    def get_cursor(
        self, schema: Schema, physical_partition_no: int
    ) -> "PartitionCursor":
        """Get :class:`.PartitionCursor` based on
        dataframe schema and physical partition number. You normally don't call
        this method directly

        :param schema: the dataframe schema this partition spec to operate on
        :param physical_partition_no: physical partition no passed in by
          :class:`~fugue.execution.execution_engine.ExecutionEngine`
        :return: PartitionCursor object
        """
        return PartitionCursor(schema, self, physical_partition_no)

    def get_partitioner(self, schema: Schema) -> SchemaedDataPartitioner:
        """Get :class:`~triad.utils.pyarrow.SchemaedDataPartitioner` by input
        dataframe schema

        :param schema: the dataframe schema this partition spec to operate on
        :return: SchemaedDataPartitioner object
        """
        pos = [schema.index_of_key(key) for key in self.partition_by]
        return SchemaedDataPartitioner(
            schema.pa_schema,
            pos,
            sizer=None,
            row_limit=self._row_limit,
            size_limit=self._size_limit,
        )

    def _update_dict(self, d: Dict[str, Any], u: Dict[str, Any]) -> None:
        for k, v in u.items():
            if k == "by":
                k = "partition_by"
            if k == "num":
                k = "num_partitions"
            d[k] = v


class DatasetPartitionCursor:
    """The cursor pointing at the first item of each logical partition inside
    a physical partition.

    It's important to understand the concept of partition, please read
    |PartitionTutorial|

    :param physical_partition_no: physical partition number passed in by
      :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    def __init__(self, physical_partition_no: int):
        self._physical_partition_no = physical_partition_no
        # the following will be set by the framework
        self._partition_no = 0
        self._slice_no = 0

    def set(self, item: Any, partition_no: int, slice_no: int) -> None:
        """reset the cursor to a row (which should be the first row of a
        new logical partition)

        :param item: an item of the dataset, or an function generating the item
        :param partition_no: logical partition number
        :param slice_no: slice number inside the logical partition (to be deprecated)
        """
        self._item = item
        self._partition_no = partition_no
        self._slice_no = slice_no

    @property
    def item(self) -> Any:
        """Get current item"""
        if callable(self._item):
            self._item = self._item()
        return self._item

    @property
    def partition_no(self) -> int:
        """Logical partition number"""
        return self._partition_no

    @property
    def physical_partition_no(self) -> int:
        """Physical partition number"""
        return self._physical_partition_no

    @property
    def slice_no(self) -> int:
        """Slice number (inside the current logical partition), for now
        it should always be 0
        """
        return self._slice_no


class BagPartitionCursor(DatasetPartitionCursor):
    """The cursor pointing at the first bag item of each logical partition inside
    a physical partition.

    It's important to understand the concept of partition, please read
    |PartitionTutorial|

    :param physical_partition_no: physical partition number passed in by
      :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    pass


class PartitionCursor(DatasetPartitionCursor):
    """The cursor pointing at the first row of each logical partition inside
    a physical partition.

    It's important to understand the concept of partition, please read
    |PartitionTutorial|

    :param schema: input dataframe schema
    :param spec: partition spec
    :param physical_partition_no: physical partition number passed in by
      :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    def __init__(self, schema: Schema, spec: PartitionSpec, physical_partition_no: int):
        self._orig_schema = schema
        self._key_index = [schema.index_of_key(key) for key in spec.partition_by]
        self._schema = schema.extract(spec.partition_by)
        self._physical_partition_no = physical_partition_no
        super().__init__(physical_partition_no=physical_partition_no)

    def set(self, row: Any, partition_no: int, slice_no: int) -> None:
        """reset the cursor to a row (which should be the first row of a
        new logical partition)

        :param row: list-like row data or a function generating a list-like row
        :param partition_no: logical partition number
        :param slice_no: slice number inside the logical partition (to be deprecated)
        """
        super().set(
            list(row) if not callable(row) else lambda: list(row()),
            partition_no=partition_no,
            slice_no=slice_no,
        )

    @property
    def row(self) -> List[Any]:
        """Get current row data"""
        return self.item

    @property
    def row_schema(self) -> Schema:
        """Schema of the current row"""
        return self._orig_schema

    @property
    def key_schema(self) -> Schema:
        """Partition key schema"""
        return self._schema

    @property
    def key_value_dict(self) -> Dict[str, Any]:
        """Based on current row, get the partition key values as a dict"""
        return {self.row_schema.names[i]: self.row[i] for i in self._key_index}

    @property
    def key_value_array(self) -> List[Any]:
        """Based on current row, get the partition key values as an array"""
        return [self.row[i] for i in self._key_index]

    def __getitem__(self, key: str) -> Any:
        """Get value by key from the current row

        :param key: column name
        :return: value in the column
        """
        return self.row[self.row_schema.index_of_key(key)]



# flake8: noqa
# pylint: disable-all
from fugue.collections.sql import transpile_sql
from fugue.dataframe import (
    alter_columns,
    as_array,
    as_array_iterable,
    as_arrow,
    as_dict_iterable,
    as_dicts,
    as_pandas,
    drop_columns,
    fugue_annotated_param,
    get_column_names,
    get_schema,
    head,
    is_df,
    peek_array,
    peek_dict,
    rename,
    select_columns,
)
from fugue.dataset import (
    as_fugue_dataset,
    as_local,
    as_local_bounded,
    count,
    get_dataset_display,
    get_num_partitions,
    is_bounded,
    is_empty,
    is_local,
)
from fugue.execution.api import as_fugue_engine_df
from fugue.execution.factory import (
    infer_execution_engine,
    parse_execution_engine,
    parse_sql_engine,
)
from fugue.extensions.creator import parse_creator
from fugue.extensions.outputter import parse_outputter
from fugue.extensions.processor import parse_processor
from fugue.extensions.transformer import parse_output_transformer, parse_transformer



from types import FrameType, TracebackType
from typing import Callable, List, Optional

_MODIFIED_EXCEPTION_VAR_NAME = "__modified_exception__"


def frames_to_traceback(
    frame: Optional[FrameType],
    limit: int,
    should_prune: Optional[Callable[[str], bool]] = None,
) -> Optional[TracebackType]:
    ctb: Optional[TracebackType] = None
    skipped = False
    while frame is not None and limit > 0:
        if _MODIFIED_EXCEPTION_VAR_NAME in frame.f_locals:
            return TracebackType(
                tb_next=None,
                tb_frame=frame,
                tb_lasti=frame.f_lasti,
                tb_lineno=frame.f_lineno,
            )
        if not skipped:
            if should_prune is not None and should_prune(frame.f_globals["__name__"]):
                frame = frame.f_back
                continue
            skipped = True
        if should_prune is None or not should_prune(frame.f_globals["__name__"]):
            ctb = TracebackType(
                tb_next=ctb,
                tb_frame=frame,
                tb_lasti=frame.f_lasti,
                tb_lineno=frame.f_lineno,
            )
            limit -= 1
            frame = frame.f_back
            continue
        break  # pragma: no cover

    return ctb


def modify_traceback(
    traceback: Optional[TracebackType],
    should_prune: Optional[Callable[[str], bool]] = None,
    add_traceback: Optional[TracebackType] = None,
) -> Optional[TracebackType]:
    ctb: Optional[TracebackType] = None

    # get stack
    stack: List[TracebackType] = []

    if add_traceback is not None:
        f: Optional[TracebackType] = add_traceback
        while f is not None:
            stack.append(f)
            f = f.tb_next
    f = traceback
    while f is not None:
        stack.append(f)
        f = f.tb_next
    stack.reverse()

    # prune and reconstruct
    for n, f in enumerate(stack):
        if (
            n == 0
            or should_prune is None
            or not should_prune(f.tb_frame.f_globals["__name__"])
        ):
            ctb = TracebackType(
                tb_next=ctb,
                tb_frame=f.tb_frame,
                tb_lasti=f.tb_lasti,
                tb_lineno=f.tb_lineno,
            )

    return ctb



from typing import Any, Callable, Dict, List, Optional, Type, Union

import pyarrow as pa
import ray
from duckdb import DuckDBPyConnection
from packaging import version
from triad import Schema, assert_or_throw, to_uuid
from triad.utils.threading import RunOnce

from fugue import (
    ArrowDataFrame,
    DataFrame,
    ExecutionEngine,
    LocalDataFrame,
    MapEngine,
    PartitionCursor,
    PartitionSpec,
)
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from fugue_duckdb.dataframe import DuckDataFrame
from fugue_duckdb.execution_engine import DuckExecutionEngine

from ._constants import FUGUE_RAY_DEFAULT_BATCH_SIZE, FUGUE_RAY_ZERO_COPY
from ._utils.cluster import get_default_partitions, get_default_shuffle_partitions
from ._utils.dataframe import add_coarse_partition_key, add_partition_key
from ._utils.io import RayIO
from .dataframe import RayDataFrame

_RAY_PARTITION_KEY = "__ray_partition_key__"


class RayMapEngine(MapEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return RayExecutionEngine

    @property
    def is_distributed(self) -> bool:
        return True

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        if len(partition_spec.partition_by) == 0:
            return self._map(
                df=df,
                map_func=map_func,
                output_schema=output_schema,
                partition_spec=partition_spec,
                on_init=on_init,
            )
        else:
            return self._group_map(
                df=df,
                map_func=map_func,
                output_schema=output_schema,
                partition_spec=partition_spec,
                on_init=on_init,
            )

    def _group_map(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
    ) -> DataFrame:
        output_schema = Schema(output_schema)
        input_schema = df.schema
        presort = partition_spec.get_sorts(
            input_schema, with_partition_keys=partition_spec.algo == "coarse"
        )
        presort_tuples = [
            (k, "ascending" if v else "descending") for k, v in presort.items()
        ]
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _udf(adf: pa.Table) -> pa.Table:  # pragma: no cover
            if adf.shape[0] == 0:
                return output_schema.create_empty_arrow_table()
            adf = adf.remove_column(len(input_schema))  # remove partition key
            if len(partition_spec.presort) > 0:
                if version.parse(pa.__version__).major < 7:  # pragma: no cover
                    idx = pa.compute.sort_indices(
                        adf, options=pa.compute.SortOptions(presort_tuples)
                    )
                    adf = adf.take(idx)
                else:
                    adf = adf.sort_by(presort_tuples)
            input_df = ArrowDataFrame(adf)
            if on_init_once is not None:
                on_init_once(0, input_df)
            cursor.set(lambda: input_df.peek_array(), 0, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_arrow()

        _df: RayDataFrame = self.execution_engine._to_ray_df(df)  # type: ignore
        if partition_spec.num_partitions != "0":
            _df = self.execution_engine.repartition(_df, partition_spec)  # type: ignore
        else:
            n = get_default_shuffle_partitions(self.execution_engine)
            if n > 0 and n != _df.num_partitions:
                # if n==0 or same as the current dataframe partitions
                # then no repartition will be done by fugue
                # otherwise, repartition the dataset
                _df = self.execution_engine.repartition(  # type: ignore
                    _df, PartitionSpec(num=n)
                )
        if partition_spec.algo != "coarse":
            rdf, _ = add_partition_key(
                _df.native,
                keys=partition_spec.partition_by,
                input_schema=input_schema,
                output_key=_RAY_PARTITION_KEY,
            )
        else:
            rdf = add_coarse_partition_key(
                _df.native,
                keys=partition_spec.partition_by,
                output_key=_RAY_PARTITION_KEY,
                bucket=_df.num_partitions,
            )

        gdf = rdf.groupby(_RAY_PARTITION_KEY)
        sdf = gdf.map_groups(
            _udf,
            batch_format="pyarrow",
            **self.execution_engine._get_remote_args(),  # type: ignore
        )
        return RayDataFrame(sdf, schema=output_schema, internal_schema=True)

    def _map(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
    ) -> DataFrame:
        output_schema = Schema(output_schema)
        input_schema = df.schema
        cursor = partition_spec.get_cursor(input_schema, 0)
        on_init_once: Any = (
            None
            if on_init is None
            else RunOnce(
                on_init, lambda *args, **kwargs: to_uuid(id(on_init), id(args[0]))
            )
        )

        def _udf(adf: pa.Table) -> pa.Table:  # pragma: no cover
            if adf.shape[0] == 0:
                return output_schema.create_empty_arrow_table()
            input_df = ArrowDataFrame(adf)
            if on_init_once is not None:
                on_init_once(0, input_df)
            cursor.set(lambda: input_df.peek_array(), 0, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_arrow()

        rdf = self.execution_engine._to_ray_df(df)  # type: ignore
        if not partition_spec.empty:
            rdf = self.execution_engine.repartition(  # type: ignore
                rdf, partition_spec=partition_spec
            )
        elif rdf.num_partitions <= 1:
            n = get_default_partitions(self.execution_engine)
            if n > 0 and n != rdf.num_partitions:
                # if n==0 or same as the current dataframe partitions
                # then no repartition will be done by fugue
                # otherwise, repartition the dataset
                rdf = self.execution_engine.repartition(  # type: ignore
                    rdf, PartitionSpec(num=n)
                )
        mb_args: Dict[str, Any] = {}
        if FUGUE_RAY_DEFAULT_BATCH_SIZE in self.conf:
            mb_args["batch_size"] = self.conf.get_or_throw(
                FUGUE_RAY_DEFAULT_BATCH_SIZE, int
            )
        if ray.__version__ >= "2.3":
            mb_args["zero_copy_batch"] = self.conf.get(FUGUE_RAY_ZERO_COPY, True)
        sdf = rdf.native.map_batches(
            _udf,
            batch_format="pyarrow",
            **mb_args,
            **self.execution_engine._get_remote_args(),  # type: ignore
        )
        return RayDataFrame(sdf, schema=output_schema, internal_schema=True)


class RayExecutionEngine(DuckExecutionEngine):
    """A hybrid engine of Ray and DuckDB as Phase 1 of Fugue Ray integration.
    Most operations will be done by DuckDB, but for ``map``, it will use Ray.

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    :param connection: DuckDB connection
    """

    def __init__(
        self, conf: Any = None, connection: Optional[DuckDBPyConnection] = None
    ):
        if not ray.is_initialized():  # pragma: no cover
            ray.init()

        super().__init__(conf, connection)
        self._io = RayIO(self)

    def __repr__(self) -> str:
        return "RayExecutionEngine"

    @property
    def is_distributed(self) -> bool:
        return True

    def create_default_map_engine(self) -> MapEngine:
        return RayMapEngine(self)

    def get_current_parallelism(self) -> int:
        res = ray.cluster_resources()
        n = res.get("CPU", 0)
        if n == 0:  # pragma: no cover
            res.get("cpu", 0)
        return int(n)

    def to_df(self, df: Any, schema: Any = None) -> DataFrame:
        return self._to_ray_df(df, schema=schema)

    def repartition(self, df: DataFrame, partition_spec: PartitionSpec) -> DataFrame:
        def _persist_and_count(df: RayDataFrame) -> int:
            self.persist(df)
            return df.count()

        rdf = self._to_ray_df(df)

        num_funcs = {
            KEYWORD_ROWCOUNT: lambda: _persist_and_count(rdf),
            KEYWORD_PARALLELISM: lambda: self.get_current_parallelism(),
        }
        num = partition_spec.get_num_partitions(**num_funcs)
        pdf = rdf.native

        if num > 0:
            if partition_spec.algo in ["default", "hash", "even", "coarse"]:
                pdf = pdf.repartition(num)
            elif partition_spec.algo == "rand":
                pdf = pdf.repartition(num, shuffle=True)
            else:  # pragma: no cover
                raise NotImplementedError(partition_spec.algo + " is not supported")
        return RayDataFrame(pdf, schema=rdf.schema, internal_schema=True)

    def broadcast(self, df: DataFrame) -> DataFrame:
        return df

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        df = self._to_auto_df(df)
        if isinstance(df, RayDataFrame):
            return df.persist(**kwargs)
        return df  # pragma: no cover

    def convert_yield_dataframe(self, df: DataFrame, as_local: bool) -> DataFrame:
        if isinstance(df, RayDataFrame):
            return df if not as_local else df.as_local()
        return super().convert_yield_dataframe(df, as_local)

    def union(self, df1: DataFrame, df2: DataFrame, distinct: bool = True) -> DataFrame:
        if distinct:
            return super().union(df1, df2, distinct)
        assert_or_throw(
            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
        )
        tdf1 = self._to_ray_df(df1)
        tdf2 = self._to_ray_df(df2)
        return RayDataFrame(tdf1.native.union(tdf2.native), df1.schema)

    def load_df(  # type:ignore
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DataFrame:
        return self._io.load_df(
            uri=path, format_hint=format_hint, columns=columns, **kwargs
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        df = self._to_ray_df(df)
        self._io.save_df(
            df,
            uri=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=partition_spec,
            force_single=force_single,
            **kwargs,
        )

    def _to_ray_df(self, df: Any, schema: Any = None) -> RayDataFrame:
        # TODO: remove this in phase 2
        res = self._to_auto_df(df, schema)
        if not isinstance(res, RayDataFrame):
            return RayDataFrame(res)
        return res

    def _to_auto_df(self, df: Any, schema: Any = None) -> DataFrame:
        # TODO: remove this in phase 2
        if isinstance(df, (DuckDataFrame, RayDataFrame)):
            assert_or_throw(
                schema is None,
                ValueError("schema must be None when df is a DataFrame"),
            )
            return df
        return RayDataFrame(df, schema)

    def _get_remote_args(self) -> Dict[str, Any]:
        res: Dict[str, Any] = {}

        for k, v in self.conf.items():
            if k.startswith("fugue.ray.remote."):
                key = k.split(".", 3)[-1]
                res[key] = v

        return res
